//+------------------------------------------------------------------+
//|                                                           TM.mqh |
//|                                       Copyright 2010, KTS Group. |
//|                                               http://www.koss.su |
//+------------------------------------------------------------------+
#property copyright "Copyright 2010, KTS Group."
#property link      "http://www.koss.su"

#include <Trade\PositionInfo.mqh>
#include <Trade\SymbolInfo.mqh>

#define  Loss   0
#define  Take   1
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CTradingManager
  {
protected:
   CPositionInfo     PosInfo;
   CSymbolInfo       SymbInfo;
   string            Name;
   int               digits;
   int               Spread;
   double            point;
   int               StopLevel;
   int               FreezeLevel;
   double            Bid;
   double            Ask;
protected:
   ushort            CheckStops(ushort Step);
   bool              UpdateSymbolInfo(string Symb);
   double            GetIndicatorValue(int handle);
   double            PartLot(string Symb,double CurrentLot,uchar VolumePercent,uint &IterationCount,double &rests);
   bool              PositionIsFrozen(ENUM_POSITION_TYPE OPEN_POSITION,double SL,double TP);
   bool              StopsIsInvalid(ENUM_POSITION_TYPE OPEN_POSITION,double Price,uchar OrderType=Loss);
   bool              StopsIsInvalid(ENUM_ORDER_TYPE OPEN_ORDER,double Price,uchar OrderType=Loss);
   bool              SendRequest(bool &ContinueRequest);
protected:
   MqlTradeRequest   Trade_request;
   MqlTradeResult    Trade_result;
   MqlTradeCheckResult Check_result;
   ulong             m_deviation;
   ulong             m_magic;
public:
   void              SimpleTrailing(string Symb,int BuyStep,int SellStep,int Deviation=0,ulong Magic=0);
   void              TATrailing(int handle,string Symb,ENUM_TIMEFRAMES SFrame=PERIOD_CURRENT,ulong Magic=0);
   bool              ModifyPosition(string Symb,double SL=0,double TP=0,ulong Magic=0);
   bool              OpenPosition(string Symb,ENUM_ORDER_TYPE Type,float Lot,double SL=0,double TP=0,string Comm="");
   void              SetTraderInfo(ulong mn,ulong Slippage=1);
   bool              CheckOpenPositions(string Symb,ENUM_POSITION_TYPE &Type,ulong Magic=0);
   bool              ClosePosition(string Symb,uchar VolumePercent,ulong deviation=ULONG_MAX);
  };
//+------------------------------------------------------------------+
//|               |
//+------------------------------------------------------------------+

bool CTradingManager::UpdateSymbolInfo(string Symb)
  {
   SymbInfo.Name(Symb);
   if(SymbInfo.Select() && SymbInfo.RefreshRates())
     {
      Name       =SymbInfo.Name();
      digits     =SymbInfo.Digits();
      point      =SymbInfo.Point();
      StopLevel  =SymbInfo.StopsLevel();
      FreezeLevel=SymbInfo.FreezeLevel();
      Bid        =SymbInfo.Bid();
      Ask        =SymbInfo.Ask();
      Spread     =SymbInfo.Spread();
      return(true);
     }
   return(NULL);
  }
//+------------------------------------------------------------------+
//|   StopsLevel                         |
//+------------------------------------------------------------------+
ushort CTradingManager::CheckStops(ushort Step)
  {
   ushort new_step=ushort(StopLevel+1);
   ushort res=(Step<new_step)?new_step:Step;
   return(res);
  }
//+------------------------------------------------------------------+
//|   StopsLevel                         |
//+------------------------------------------------------------------+  

bool  CTradingManager::StopsIsInvalid(ENUM_POSITION_TYPE OPEN_POSITION,double Price,uchar OrderType=Loss)
  {
   bool IsInvalid=NULL;
   switch(OPEN_POSITION)
     {
      case POSITION_TYPE_BUY:
        {
         switch(OrderType)
           {
            case Loss:
               IsInvalid=(Price!=0 && (Bid-Price)/point<StopLevel);
               break;

            case Take:
               IsInvalid=(Price!=0 && (Price-Bid)/point<StopLevel);
               break;
           }
        }
      break;

      case POSITION_TYPE_SELL:
        {
         switch(OrderType)
           {
            case Loss:
               IsInvalid=(Price!=0 && (Price-Ask)/point<StopLevel);
               break;

            case Take:
               IsInvalid=(Price!=0 && (Ask-Price)/point<StopLevel);
               break;
           }
        }
      break;
     }
   return(IsInvalid);
  }
//+------------------------------------------------------------------+
//|   StopsLevel                         |
//+------------------------------------------------------------------+  

bool  CTradingManager::StopsIsInvalid(ENUM_ORDER_TYPE OPEN_ORDER,double Price,uchar OrderType=Loss)
  {
   bool IsInvalid=NULL;
   switch(OPEN_ORDER)
     {
      case ORDER_TYPE_BUY:
        {
         switch(OrderType)
           {
            case Loss:
               IsInvalid=(Price!=0 && (Bid-Price)/point<StopLevel);
               break;

            case Take:
               IsInvalid=(Price!=0 && (Price-Bid)/point<StopLevel);
               break;
           }
        }
      break;

      case ORDER_TYPE_SELL:
        {
         switch(OrderType)
           {
            case Loss:
               IsInvalid=(Price!=0 && (Price-Ask)/point<StopLevel);
               break;

            case Take:
               IsInvalid=(Price!=0 && (Ask-Price)/point<StopLevel);
               break;
           }
        }
      break;
     }
   return(IsInvalid);
  }
//+------------------------------------------------------------------+
//|                         |
//+------------------------------------------------------------------+

bool CTradingManager::PositionIsFrozen(ENUM_POSITION_TYPE OPEN_POSITION,double SL,double TP)
  {
   bool Frozen=NULL;
   switch(OPEN_POSITION)
     {
      case POSITION_TYPE_BUY:
         Frozen=((TP!=0 && TP-Bid<FreezeLevel) || (SL!=0 && Bid-SL<FreezeLevel));
         break;

      case POSITION_TYPE_SELL:
         Frozen=((TP!=0 && Ask-TP<FreezeLevel) || (SL!=0 && SL-Ask<FreezeLevel));
         break;
     }

   return(Frozen);
  }
//+------------------------------------------------------------------+
//| -                                        |
//+------------------------------------------------------------------+

void CTradingManager::SimpleTrailing(string Symb,int BuyStep,int SellStep,int Deviation=0,ulong Magic=0)
  {
   if(PosInfo.Select(Symb))
     {
      if(UpdateSymbolInfo(Symb))
        {
         bool   Modified=NULL;
         double sl=PosInfo.StopLoss();
         double tp=PosInfo.TakeProfit();
         double op=PosInfo.PriceOpen();
         double new_sl=NULL;

         switch(PosInfo.Type())
           {
            case POSITION_TYPE_BUY:
              {
               if(Bid-op>BuyStep*point)
                 {
                  if(sl<Bid-(BuyStep+Deviation)*point)
                    {
                     new_sl=NormalizeDouble(Bid-BuyStep*point,digits);
                     Modified=(!PositionIsFrozen(POSITION_TYPE_BUY,sl,tp) && !StopsIsInvalid(POSITION_TYPE_BUY,new_sl,Loss));
                    }
                 }
              }
            break;

            case POSITION_TYPE_SELL:
              {
               if(op-Ask>SellStep*point)
                 {
                  if(sl>Ask+(SellStep+Deviation)*point)
                    {
                     new_sl=NormalizeDouble(Ask+SellStep*point,digits);
                     Modified=(!PositionIsFrozen(POSITION_TYPE_SELL,sl,tp) && !StopsIsInvalid(POSITION_TYPE_SELL,new_sl,Loss));
                    }
                 }
              }
            break;
           }

         if(Modified && ModifyPosition(Name,new_sl,tp,Magic))
           {
            Print("StopLoss was modified-",DoubleToString(new_sl,digits));
           }
        }
     }
  }
//+------------------------------------------------------------------+
//|                           |
//+------------------------------------------------------------------+

double CTradingManager::GetIndicatorValue(int handle)
  {
   double buffer[];
   if(handle!=INVALID_HANDLE)
     {
      if(CopyBuffer(handle,0,1,1,buffer)<0) return(0);
      else return(buffer[0]);
     }
   return(0);
  }
//+------------------------------------------------------------------+
//| -                                        |
//+------------------------------------------------------------------+

void CTradingManager::TATrailing(int handle,string Symb,ENUM_TIMEFRAMES SFrame=PERIOD_CURRENT,ulong Magic=0)
  {
   if(PosInfo.Select(Symb))
     {
      if(UpdateSymbolInfo(Symb))
        {
         double new_sl=NormalizeDouble(GetIndicatorValue(handle),digits);
         if(new_sl!=0)
           {
            double High[];
            double Low[];
            double tp=PosInfo.TakeProfit();
            double sl=PosInfo.StopLoss();
            bool   Modified=NULL;

            switch(PosInfo.Type())
              {
               case POSITION_TYPE_BUY:
                  if(CopyLow(Name,SFrame,1,1,Low)>0)
                    {
                     if(new_sl<Low[0] && sl>0 && new_sl>sl)
                       {
                        Modified=(!StopsIsInvalid(POSITION_TYPE_BUY,new_sl,Loss) && !PositionIsFrozen(POSITION_TYPE_BUY,sl,tp));
                       }
                    }
                  break;

               case POSITION_TYPE_SELL:
                  if(CopyHigh(Name,SFrame,1,1,High)>0)
                    {
                     if(new_sl>High[0] && sl>0 && new_sl<sl)
                       {
                        Modified=(!StopsIsInvalid(POSITION_TYPE_SELL,new_sl,Loss) && !PositionIsFrozen(POSITION_TYPE_SELL,sl,tp));
                       }
                    }
                  break;
              }

            if(Modified && ModifyPosition(Name,new_sl,tp,Magic))
              {
               Print("StopLoss was modified-",DoubleToString(new_sl,digits));
              }
           }
        }
     }
  }
//+------------------------------------------------------------------+
//|                          |
//+------------------------------------------------------------------+

void  CTradingManager::SetTraderInfo(ulong mn,ulong Slippage=1)
  {
   m_deviation=Slippage*Spread;
   m_magic=mn;
  }
//+------------------------------------------------------------------+
//|                                    |
//+------------------------------------------------------------------+
bool CTradingManager::CheckOpenPositions(string Symb,ENUM_POSITION_TYPE &Type,ulong Magic=0)
  {
   if(PosInfo.Select(Symb))
     {
      if(PosInfo.Magic()==Magic || Magic==0)
        {
         Type=PosInfo.Type();
         return(true);
        }
     }
   return(NULL);
  }
//+------------------------------------------------------------------+
//|                                        |
//+------------------------------------------------------------------+

bool CTradingManager::ModifyPosition(string Symb,double SL=0,double TP=0,ulong Magic=0)
  {
   bool Done=NULL;

   if(PosInfo.Select(Symb))
     {
      if(PosInfo.Magic()==Magic || Magic==0)
        {
         do
           {
            bool RequestDone=NULL;
            bool Continue=NULL;
            
            Trade_request.action=TRADE_ACTION_SLTP;
            Trade_request.symbol=Symb;
            Trade_request.sl    =SL;
            Trade_request.tp    =TP;
            
            RequestDone=SendRequest(Continue);
            if(RequestDone) Done=true;
             else
              {
               if(Continue) Done=NULL;
                else return(NULL);
              }
           }
         while(!Done);
        }
     }
   return(Done);
  }
//+------------------------------------------------------------------+
//|                                           |
//+------------------------------------------------------------------+

bool CTradingManager::OpenPosition(string Symb,ENUM_ORDER_TYPE Type,float Lot,double SL=0,double TP=0,string Comm="")
  {
   bool   Done=NULL;
   double price=NULL;

   if(Type!=ORDER_TYPE_BUY && Type!=ORDER_TYPE_SELL) return(Done);

   if(UpdateSymbolInfo(Symb))
     {
      //--- checking
      if(!StopsIsInvalid(Type,SL,Loss) && !StopsIsInvalid(Type,TP,Take))
        {
         do
           {
            bool RequestDone=NULL;
            bool Continue=NULL;

            //--- setting request
            Trade_request.action      =TRADE_ACTION_DEAL;
            Trade_request.symbol      =Name;
            Trade_request.magic       =m_magic;
            Trade_request.volume      =Lot;
            Trade_request.type        =Type;
            Trade_request.price       =(Type==ORDER_TYPE_BUY) ? Ask : Bid;
            Trade_request.sl          =SL;
            Trade_request.tp          =TP;
            Trade_request.deviation   =m_deviation;
            Trade_request.type_filling=ORDER_FILLING_AON;
            Trade_request.comment     =Comm;
            //---
            RequestDone=SendRequest(Continue);

            if(RequestDone) Done=true;
            else
              {
               if(Continue) Done=false;
               else return(false);
              }
           }
         while(!Done);
        }
     }
   return(Done);
  }
//+------------------------------------------------------------------+
//|                           |
//+------------------------------------------------------------------+

double  CTradingManager::PartLot(string Symb,double Lot,uchar VolumePercent,uint &IterationCount,double &rests)
  {
   double LotMin=SymbolInfoDouble(Symb,SYMBOL_VOLUME_MIN);
   double LotMax=SymbolInfoDouble(Symb,SYMBOL_VOLUME_MAX);
   double LotStep=SymbolInfoDouble(Symb,SYMBOL_VOLUME_STEP);

   if(Lot==LotMin) return(Lot);

   double k=NULL;
   double NeedVolume=NULL;
   double part=NULL;
   uchar  Percent=VolumePercent;

   if(Percent==0 || Percent>100) Percent=100;      //  

   if(LotMin==0) LotMin=0.1;
   if(LotMax==0) LotMax=100;
   if(LotStep>0) k=1/LotStep; else k=1/LotMin;

   part=(Lot*Percent)/100;
   NeedVolume=MathFloor(part*k)/k;

   if(NeedVolume<LotMin || Lot-NeedVolume<LotMin) return(Lot);

   if(NeedVolume>LotMax)
     {
      rests=NeedVolume-LotMax;
      IterationCount=(int)MathCeil(NeedVolume/LotMax);
      Print("Position on "+Symb+" will be closed by parts.");
      return(LotMax);
     }
   return(NeedVolume);
  }
//+------------------------------------------------------------------+
//|                                                   |
//+------------------------------------------------------------------+

bool CTradingManager::ClosePosition(string Symb,uchar VolumePercent,ulong deviation=ULONG_MAX)
  {

   bool   Closed=NULL;
   uint   i=NULL;

   if(PosInfo.Select(Symb) && UpdateSymbolInfo(Symb))
     {
      if(!PositionIsFrozen(PosInfo.Type(),PosInfo.StopLoss(),PosInfo.TakeProfit()))
        {
         double VolumeForClose=NULL;
         double CurrentVolume=PosInfo.Volume();
         static double rests=NULL;                  //  ,      SYMBOL_VOLUME_MAX
         uint   attempts=NULL;                      //    
         bool   RequestDone=NULL;
         bool   Continue=NULL;
         do
           {
            VolumeForClose=(rests==0) ? PartLot(Symb,CurrentVolume,VolumePercent,i,rests) : PartLot(Symb,rests,100,i,rests);
            if(PosInfo.Type()==POSITION_TYPE_BUY)
              {
               Trade_request.type=ORDER_TYPE_SELL;
               Trade_request.price=Bid;
              }
            else
              {
               //--- prepare query for close SELL position
               Trade_request.type =ORDER_TYPE_BUY;
               Trade_request.price=Ask;
              }
            //--- setting request
            Trade_request.action       =TRADE_ACTION_DEAL;
            Trade_request.symbol       =Name;
            Trade_request.volume       =VolumeForClose;
            Trade_request.sl           =0.0;
            Trade_request.tp           =0.0;
            Trade_request.deviation    =(deviation==ULONG_MAX) ? m_deviation : deviation;
            Trade_request.type_filling =ORDER_FILLING_AON;
            //---
            RequestDone=SendRequest(Continue);
            if(RequestDone)
              {
               if(i>0)
                 {
                  i--;
                  attempts++;
                 }
               Closed=true;
              }
            else
              {
               if(Continue)
                 {
                  if(i>0)
                    {
                     if(attempts==0) rests=NULL;
                    }
                  else i++;
                 }
               else break;
              }
           }
         while(i>0);
        }
     }
   return(Closed);
  }
//+------------------------------------------------------------------+
//|                                   |
//+------------------------------------------------------------------+

bool CTradingManager::SendRequest(bool &ContinueRequest)
  {
   uint   CheckRetCode=NULL;
   bool   CheckRequest=NULL;
   bool   Done        =NULL;

   CheckRequest=OrderCheck(Trade_request,Check_result);
   if(CheckRequest)
     {
      OrderSend(Trade_request,Trade_result);
      switch(Trade_result.retcode)
        {
         //---  
         case TRADE_RETCODE_PLACED:
         case TRADE_RETCODE_DONE:
         case TRADE_RETCODE_DONE_PARTIAL:
           {
            Done=true;
           }
         break;

         //---
         case TRADE_RETCODE_REQUOTE:
           {
            Bid=Trade_result.bid;
            Ask=Trade_result.ask;
            ContinueRequest=true;
           }
         break;
        }
     }
   return(Done);
  }
//+------------------------------------------------------------------+
